
## Copyright (c) 2001-2010, Scott D. Peckham
## January 2009  (converted from IDL)
## July, August 2009
## May 2010 (changes to unit_test() and read_cfg_file()

#-----------------------------------------------------------------------
#  NOTES:  This file defines a Degree-Day snowmelt component
#          and related functions.  It inherits from the snowmelt
#          "base class" in "snow_base.py".
#-----------------------------------------------------------------------
#
#  unit_test()
#
#  class snow_degree_day
#
#      get_attribute()          # (10/26/11)
#      check_input_types()
#      update_meltrate()
#
#-----------------------------------------------------------------------

from numpy import *
import numpy

import snow_base
import tf_utils

#-----------------------------------------------------------------------
def unit_test():

    c = snow_component()
    c.CCA   = False
    c.DEBUG = False
    
    #-----------------------------------------
    # This function adjusts for the platform
    # and can be changed in "tf_utils.py".
    #-----------------------------------------
    cfg_directory = tf_utils.TF_Test_Directory()
    cfg_prefix    = tf_utils.TF_Test_Case_Prefix()
    c.site_prefix = 'Treynor'  #############    

    #-------------------------------------------
    # Call initialize() and call update() once
    #-------------------------------------------
##    print 'STATUS =', c.get_status()
##    c.initialize( mode="driver" )
##    print 'STATUS =', c.get_status()
##    time_sec = float64(0)
##    c.update(time_sec)
##    print 'STATUS =', c.get_status()

    c.run_model( cfg_directory=cfg_directory,
                 cfg_prefix=cfg_prefix )
    
#   unit_test()
#-----------------------------------------------------------------------
class snow_component(snow_base.snow_component):

    #-------------------------------------------------------------------
    def get_attribute(self, att_name):

        map = {'comp_name':          'SnowDegreeDay',
               'version':            '3.1',
               'model_name':         'Snowmelt_Degree_Day',    # (Snowmelt ?)
               'model_family':       'TopoFlow',
               'cfg_template_file':  'Snow_Degree_Day.cfg.in',
               'cfg_extension':      '_snow_degree_day.cfg',
               'cmt_var_prefix':     '/SnowDegreeDay/Input/Var/',
               'gui_xml_file':       '/home/csdms/cca/topoflow/3.1/src/share/cmt/gui/Snow_Degree_Day.xml',
               'dialog_title':       'Snowmelt: Degree-Day Parameters',    # (Snowmelt ?)
               'time_step_type':     'fixed',
               'time_units':         'seconds',
               'mesh_type':          'none',
               'author_name':        'Scott Peckham'}

        try:
            return map[ att_name.lower() ]
        except:
            print '###################################################'
            print ' ERROR: Could not find attribute: ' + att_name
            print '###################################################'
            print ' '

    #   get_attribute() 
    #-------------------------------------------------------------------
    def check_input_types(self):

        #--------------------------------------------------
        # Notes: rho_H2O, Cp_snow, rho_air and Cp_air are
        #        currently always scalars.
        #--------------------------------------------------        
        are_scalars = numpy.array([
##                          self.pp.is_scalar('rate'),   # (not "rates")
##                          self.pp.is_scalar('duration'),
                          #-------------------------------
                          self.mp.is_scalar('P'),
                          self.mp.is_scalar('rho_H2O'),
                          self.mp.is_scalar('rho_air'),
                          self.mp.is_scalar('T_air'),
                          self.mp.is_scalar('Cp_air'),
                          #-------------------------------
                          self.is_scalar('rho_snow'),
                          self.is_scalar('Cp_snow'),
                          self.is_scalar('h0_snow'),
                          self.is_scalar('h0_swe'),
                          self.is_scalar('c0'),
                          self.is_scalar('T0') ])

        self.ALL_SCALARS = numpy.all(are_scalars)
        
    #   check_input_types()
    #-------------------------------------------------------------------
    def update_meltrate(self):

        #------------------------------------------------------------
        #Notes:  Arguments are assumed to be references to a scalar
        #        or grid so that:
        #        c0     = degree-day melt factor [mm/day/deg_C]
        #        T0     = threshold temperature [deg_C]
        #        M      = water equivalent of snowmelt [m/s]
        #        M_max  = max possible meltrate if all snow melts
        #        T_air  = air temperature [deg_C]

        #        Model must start when snow is isothermal.
        #        Cooling of the snowpack is not considered.

        #        This is a simple snowmelt model that can be used
        #        when there is insufficient data to use the energy
        #        balance method.  If c0 and T0 are established for a
        #        range of conditions, Kane et al. (1997) showed that
        #        this method gives comparable results to the energy
        #        balance method.

        #        86400000d = 1000 [mm/m] * 60 [sec/min] *
        #                    60 [min/hr] * 24 [hrs/day]

        #        rho_snow is not needed in formula here, but is
        #        needed to convert snowmelt to snow water
        #        equivalent (swe) in update_swe() in "snow_base.py".
        #-------------------------------------------------------------
        T_air = self.get_port_data('T_air', self.mp)
        
        #------------------------------
        # Compute degree-day meltrate
        #------------------------------
        M = (self.c0 / float64(8.64E7)) * (T_air - self.T0)   #[m/s]

        # This is really an "enforce_min_meltrate()"
        self.SM = maximum(M, float64(0))
   
        #-------------------------------------------------------
        # Note: enforce_max_meltrate() method is always called
        #       by the base class to make sure that meltrate
        #       does not exceed the max possible.
        #-------------------------------------------------------
        #  self.enforce_max_meltrate()
        
    #   update_meltrate()
    #-------------------------------------------------------------------
    
